@@ -23,7 +23,8 @@
 // re-cast an io_pending_t into this more descriptive structure.
 // the first few items _must_ match the original struct.
 typedef struct _io_pending_storage_t {
-    io_queue_t *q;
+    int io_queue_type;
+    LIBEVENT_THREAD *thread;
     conn *c;
     mc_resp *resp;            /* original struct ends here */
     item *hdr_it;             /* original header item. */
@@ -69,6 +70,9 @@ void process_extstore_stats(ADD_STAT add_stats, conn *c) {
     assert(add_stats);
 
     void *storage = c->thread->storage;
+    if (storage == NULL) {
+        return;
+    }
     extstore_get_stats(storage, &st);
     st.page_data = calloc(st.page_count, sizeof(struct extstore_page_data));
     extstore_get_page_data(storage, &st);
@@ -220,13 +224,14 @@ static void _storage_get_item_cb(void *e, obj_io *io, int ret) {
         p->miss = false;
     }
 
-    p->q->count--;
     p->active = false;
     //assert(c->io_wrapleft >= 0);
 
     // All IO's have returned, lets re-attach this connection to our original
     // thread.
-    if (p->q->count == 0) {
+    io_queue_t *q = conn_io_queue_get(p->c, p->io_queue_type);
+    q->count--;
+    if (q->count == 0) {
         redispatch_conn(c);
     }
 }
@@ -269,7 +274,7 @@ int storage_get_item(conn *c, item *it, mc_resp *resp) {
     // io_pending owns the reference for this object now.
     p->hdr_it = it;
     p->resp = resp;
-    p->q = q; // quicker access to the queue structure.
+    p->io_queue_type = IO_QUEUE_EXTSTORE;
     obj_io *eio = &p->io_ctx;
 
     // FIXME: error handling.
@@ -363,9 +368,9 @@ int storage_get_item(conn *c, item *it, mc_resp *resp) {
     return 0;
 }
 
-void storage_submit_cb(void *ctx, void *stack) {
+void storage_submit_cb(io_queue_t *q) {
     // Don't need to do anything special for extstore.
-    extstore_submit(ctx, stack);
+    extstore_submit(q->ctx, q->stack_ctx);
 }
 
 static void recache_or_free(io_pending_t *pending) {
@@ -374,9 +379,9 @@ static void recache_or_free(io_pending_t *pending) {
 
     conn *c = p->c;
     obj_io *io = &p->io_ctx;
+    assert(io != NULL);
     item *it = (item *)io->buf;
     assert(c != NULL);
-    assert(io != NULL);
     bool do_free = true;
     if (p->active) {
         // If request never dispatched, free the read buffer but leave the
@@ -384,8 +389,10 @@ static void recache_or_free(io_pending_t *pending) {
         do_free = false;
         size_t ntotal = ITEM_ntotal(p->hdr_it);
         slabs_free(it, ntotal, slabs_clsid(ntotal));
-        p->q->count--;
-        assert(p->q->count >= 0);
+
+        io_queue_t *q = conn_io_queue_get(c, p->io_queue_type);
+        q->count--;
+        assert(q->count >= 0);
         pthread_mutex_lock(&c->thread->stats.mutex);
         c->thread->stats.get_aborted_extstore++;
         pthread_mutex_unlock(&c->thread->stats.mutex);
@@ -443,7 +450,9 @@ static void recache_or_free(io_pending_t *pending) {
 // TODO: stubbed with a reminder: should be able to move most of the extstore
 // callback code into this code instead, executing on worker thread instead of
 // IO thread.
-void storage_complete_cb(void *ctx, void *stack_ctx) {
+void storage_complete_cb(io_queue_t *q) {
+    // need to reset the stack for next use.
+    q->stack_ctx = NULL;
     return;
 }
 
